import click
import os
import sys
from dotenv import load_dotenv
import yaml
import json
import subprocess
import threading
import time

# Load environment variables from .env file in current directory or from HOME/.env
load_dotenv()
dotenv_path = os.path.join(os.path.expanduser("~"), '.env')
if os.path.exists(dotenv_path):
    load_dotenv(dotenv_path)

#Add path to current file
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from orchestration import draw_memory_key_graph, graphbuilder_sync
from lattice_drop import fetch_echonode_data, process_echonode_data, render_echonode_data, emit_live_reports
from curating_red_stones import execute_curating_red_stones
from echonode_trace_activation import execute_echonode_trace_activation
from enriched_version_fractale_001 import execute_enriched_version_fractale_001
from redstone_writer import RedStoneWriter
import requests

API_URL = os.getenv("EH_API_URL")
EH_TOKEN = os.getenv("EH_TOKEN")

@click.command()
def scan_nodes():
    """Simulate scanning and listing nodes in the system."""
    click.echo("Scanning nodes... (placeholder for recursive node scanning)")

@click.command()
def flex():
    """Demonstrate flexible orchestration of tasks."""
    click.echo("Flexing tasks... (placeholder for flexible task orchestration)")

@click.command()
def trace_orbit():
    """Trace and visualize the orbit of data or processes."""
    click.echo("Tracing orbit... (placeholder for data/process orbit tracing)")

@click.command()
def echo_sync():
    """Synchronize data or processes across nodes."""
    click.echo("Synchronizing... (placeholder for data/process synchronization)")

@click.command()
def draw_memory_graph():
    """Print an ASCII-rendered graph of the memory keys and Arc structure."""
    draw_memory_key_graph()
    echonode_data = fetch_echonode_data()
    if (echonode_data):
        processed_data = process_echonode_data(echonode_data)
        render_echonode_data(processed_data)
        # Include delta overlays
        click.echo("Delta overlays included.")

@click.command()
def curating_red_stones(verbose: bool = False, dry_run: bool = False):
    """Visualize and structure Red Stone metadata connections."""
    if verbose:
        click.echo("Activating Curating Red Stones Lattice with verbose output...")
    if dry_run:
        click.echo("Dry run mode: No changes will be committed.")
    execute_curating_red_stones()

@click.command()
def activate_echonode_trace(verbose: bool = False, dry_run: bool = False):
    """Activate and trace EchoNode sessions."""
    if verbose:
        click.echo("Activating EchoNode Trace with verbose output...")
    if dry_run:
        click.echo("Dry run mode: No changes will be committed.")
    execute_echonode_trace_activation()

@click.command()
def enrich_fractale_version(verbose: bool = False, dry_run: bool = False):
    """Enhance and enrich the Fractale 001 version."""
    if verbose:
        click.echo("Activating Enriched Version Fractale 001 with verbose output...")
    if dry_run:
        click.echo("Dry run mode: No changes will be committed.")
    execute_enriched_version_fractale_001()

@click.command()
@click.option('--api-url', required=True, help='API URL for GraphBuilderSync')
@click.option('--token', required=True, help='Authorization token for GraphBuilderSync')
@click.option('--node-id', default=None, help='Node ID for GraphBuilderSync')
@click.option('--node-data', default=None, help='Node data for GraphBuilderSync')
@click.option('--action', type=click.Choice(['push', 'pull']), default='pull', help='Action for GraphBuilderSync')
@click.option('--narrative', is_flag=True, help='Narrative context for GraphBuilderSync. For more info, visit https://YOUR_EH_API_URL/latice/tushell')
def graphbuilder_sync_command(api_url, token, node_id, node_data, action, narrative):
    """Execute GraphBuilderSync operations."""
    if narrative:
        click.echo("Executing GraphBuilderSync with narrative context...")
    result = graphbuilder_sync(api_url, token, node_id, node_data, action)
    click.echo(result)

@click.command()
@click.option('--repo-path', required=True, help='Path to the repository')
@click.option('--commit-message', required=True, help='Commit message for encoding resonance')
def redstone_encode_resonance(repo_path, commit_message):
    """Encode recursive resonance into commits."""
    writer = RedStoneWriter(repo_path)
    writer.encode_resonance(commit_message)
    click.echo("Encoded recursive resonance into commit.")

@click.command()
@click.option('--repo-path', required=True, help='Path to the repository')
@click.option('--commit-message', required=True, help='Commit message for writing narrative diffs')
@click.option('--diffs', required=True, help='Narrative diffs to be written to commit message')
def redstone_write_narrative_diffs(repo_path, commit_message, diffs):
    """Write narrative diffs to commit messages."""
    writer = RedStoneWriter(repo_path)
    narrative_diff = writer.write_narrative_diffs(commit_message, diffs)
    click.echo(narrative_diff)

@click.command()
@click.option('--repo-path', required=True, help='Path to the repository')
@click.option('--anchors', required=True, help='Resonance anchors to be stored')
def redstone_store_resonance_anchors(repo_path, anchors):
    """Store resonance anchors in .redstone.json."""
    writer = RedStoneWriter(repo_path)
    writer.store_resonance_anchors(anchors)
    click.echo("Stored resonance anchors in .redstone.json.")

@click.command()
@click.option('--repo-path', required=True, help='Path to the repository')
@click.option('--echonode-metadata', required=True, help='EchoNode metadata to be synced')
def redstone_sync_echonode_metadata(repo_path, echonode_metadata):
    """Sync with EchoNode metadata."""
    writer = RedStoneWriter(repo_path)
    writer.sync_with_echonode_metadata(echonode_metadata)
    click.echo("Synced with EchoNode metadata.")

@click.command()
@click.option('--repo-path', required=True, help='Path to the repository')
@click.option('--redstone-score', required=True, type=int, help='RedStone score for post-commit analysis')
def redstone_post_commit_analysis(repo_path, redstone_score):
    """Support RedStone score metadata field for post-commit analysis."""
    writer = RedStoneWriter(repo_path)
    writer.post_commit_analysis(redstone_score)
    click.echo("Supported RedStone score metadata field for post-commit analysis.")

@click.command()
def echo_live_reports():
    """Emit live reports from EchoNodes tied to active narrative arcs."""
    emit_live_reports()
    # Integrate with fractalstone_protocol, redstone_protocol, and EchoMuse glyph emitters
    click.echo("Live reports integrated with fractalstone_protocol, redstone_protocol, and EchoMuse glyph emitters.")

@click.command()
@click.option('--trace-id', required=False, help='Trace ID for visual replay')
@click.option('--render', is_flag=True, help='Render visual trace replay')
@click.option('--muse-mode', is_flag=True, help='Enable Muse Mode for glyph-enhanced, poetic status report')
@click.option('--init', is_flag=True, help='Initialize Muse Mode YAML dataset')
@click.option('--interactive', is_flag=True, help='Interactive mode with terminal menu choices')
@click.option('--help', is_flag=True, help='Show detailed explanation of Muse Mode')
def mia_status(trace_id, render, muse_mode, init, interactive, help):
    """Provide information about Mia's current state or status."""
    if help:
        click.echo("""
        --muse-mode: Enable Muse Mode for glyph-enhanced, poetic status report.
        Muse Mode reflects the emotional state, recent recursion activity, and redstone modulation.

        Features and Corresponding Keys:
        - Emotional State: Retrieves 'emotional_state' from the YAML file and maps to get-memory API.
        - Recursion Activity: Retrieves 'recursion_activity' from the YAML file and maps to get-memory API.
        - Redstone Modulation: Retrieves 'redstone_modulation' from the YAML file and maps to get-memory API.
        - Glyph Keys: Retrieves 'glyph_keys' from the YAML file and maps to get-memory API.
        - Echo Node Bindings: Retrieves 'echo_node_bindings' from the YAML file and maps to get-memory API.
        - Vault Whisper: Retrieves 'vault_whisper' from the YAML file and maps to get-memory API.

        Example YAML Output:
        emotional_state:
          - redstones:vcu.CeSaReT...
        recursion_activity:
          - Trace:042a0ea2...
          - Trace:072e28a3...
        redstone_modulation:
          - redstones:vcu.CeSaReT.jgwill.tushell.42...
        glyph_keys:
          - glyphstyle::SignalSet.StandardV1
        echo_node_bindings:
          - tushell_langfuse:EchoMuse.CanvasTraceSync.V1
        vault_whisper:
          - Portal:MietteTale

        Use --init to generate a YAML scaffold representing the current emotional, recursive, and narrative environment of EchoMuse.
        """)
        return

    if init:
        yaml_data = {
            "emotional_state": ["muse:emotional_state"],
            "recursion_activity": ["muse:recursion_activity"],
            "redstone_modulation": ["muse:redstone_modulation"],
            "glyph_keys": ["muse:glyph_keys"],
            "echo_node_bindings": ["muse:echo_node_bindings"],
            "vault_whisper": ["muse:vault_whisper"]
        }
        with open("muse_mode.yaml", "w") as file:
            yaml.dump(yaml_data, file)
        click.echo("Muse Mode YAML dataset initialized with key names.")
        return

    echonode_data = fetch_echonode_data()
    if not echonode_data:
        print("Error: Unable to fetch EchoNode data.- Not implemented yet.")
        print("For now, we set a dummy value so it would process the Mia-Status")
        echonode_data = {
            id: "dummy",
            trace_id: "dummy"
        }
            
    if echonode_data:
        processed_data = process_echonode_data(echonode_data)
        render_echonode_data(processed_data)
        click.echo("Mia's status has been retrieved and displayed.")

        if muse_mode:
            try:
                with open("muse_mode.yaml", "r") as file:
                    muse_data = yaml.safe_load(file)

                click.echo("🎭 Muse Mode Active:\n")
                for key, values in muse_data.items():
                    click.echo(f"{key.capitalize()}:")
                    for value in values:
                        click.echo(f"  - {value}")

                # Map features to keys and display them
                feature_key_map = {
                    "Emotional State": "emotional_state",
                    "Recursion Activity": "recursion_activity",
                    "Redstone Modulation": "redstone_modulation",
                    "Glyph Keys": "glyph_keys",
                    "Echo Node Bindings": "echo_node_bindings",
                    "Vault Whisper": "vault_whisper"
                }

                DEBUG=False
                if DEBUG:
                    click.echo("\nFeature to Key Mapping:")
                    for feature, key in feature_key_map.items():
                        if key in muse_data:
                            click.echo(f"{feature}: {muse_data[key]}")
                        else:
                            click.echo(f"{feature}: Key not found")

                # Example: Use scan-keys to find related keys
                click.echo("\nScanning for related keys...")
                click.echo("\n  JG's Notes:  I have not removed that because it could become scanning for subkeys.")
                pattern = "Trace"
                response = requests.get(f"{API_URL}/api/scan", params={"pattern": pattern}, headers={"Authorization": f"Bearer {EH_TOKEN}"})
                if response.status_code == 200:
                    keys = response.json().get('keys', [])
                    click.echo("Related Keys:")
                    for key in keys:
                        click.echo(f"  - {key}")
                else:
                    click.echo("Error scanning keys.")

                click.echo("\n---------End of a potential scanning---------------\n")
                
                # # Retrieve and display keys from the get-memory API
                # click.echo("\nRetrieving keys from get-memory API:")
                # for feature, key in feature_key_map.items():
                #     memory_key = f"muse:{key}"
                #     response = requests.get(f"{API_URL}/api/memory", params={"key": memory_key}, headers={"Authorization": f"Bearer {EH_TOKEN}"})
                #     if response.status_code == 200:
                #         value = response.json().get(memory_key, "No value found")
                #         click.echo(f"{feature}: {value}")
                #     else:
                #         click.echo(f"{feature}: Error retrieving key {memory_key}")

                # click.echo("\n--------------------------------\n")
                # Retrieve and display values from the get-memory API for each key in the YAML file
                #click.echo("\nRetrieving values from get-memory API:")
                click.echo("\n----------------Mia.status.start🎭->>>-------")
                for key, values in muse_data.items():
                    click.echo(f"{key.capitalize()}:")
                    for memory_key in values:
                        response = requests.get(f"{API_URL}/api/memory", params={"key": memory_key}, headers={"Authorization": f"Bearer {EH_TOKEN}"})
                        if response.status_code == 200:
                            value = response.json().get("value", "No value found")
                            click.echo(f"  - {value}")
                        else:
                            click.echo(f"  - Error retrieving key {memory_key}")
                click.echo("\n----------------Mia.status.end-🎭<<--------\n")

            except FileNotFoundError:
                click.echo("Error: muse_mode.yaml not found. Please initialize Muse Mode using --init.")
            except Exception as e:
                click.echo(f"An error occurred: {e}")

        if render and trace_id:
            render_LESR_timeline(trace_id)
        if interactive:
            # Implement interactive terminal menu choices
            click.echo("Interactive mode activated.")

@click.command()
def tushell_echo():
    """Provide information about the current state or status of characters."""
    echonode_data = fetch_echonode_data()
    if echonode_data:
        processed_data = process_echonode_data(echonode_data)
        render_echonode_data(processed_data)
        click.echo("Character states have been retrieved and displayed.")

@click.command()
@click.option('--key', required=True, help='Memory key to retrieve')
@click.option('--list', 'list_keys', is_flag=True, help='List all keys (writers only)')
@click.option('--json', 'output_json', is_flag=True, help='Output the result in JSON format')
def get_memory(key, list_keys, output_json):
    """Get fractal stone memory value by key."""
    params = {"key": key}
    if list_keys:
        params["list"] = True
    headers = {"Authorization": f"Bearer {EH_TOKEN}"}
    response = requests.get(f"{API_URL}/api/memory", params=params, headers=headers)
    if response.status_code == 200:
        result = response.json()
        if output_json:
            click.echo(json.dumps(result, indent=4))
        else:
            value = result.get("value", "No value found")
            click.echo(value)
            for key, val in result.items():
                if key != "value":
                    click.echo(f"-X {key}: {val}")
    elif response.status_code == 401:
        click.echo("Unauthorized: Invalid or missing authentication token.")
    else:
        click.echo(f"Error: {response.text}")

@click.command()
@click.option('--key', required=True, help='Memory key to store')
@click.option('--value', required=True, help='Value to store')
def post_memory(key, value):
    """Store fractal stone memory value by key."""
    payload = {"key": key, "value": value}
    headers = {"Authorization": f"Bearer {EH_TOKEN}"}
    response = requests.post(f"{API_URL}/api/memory", json=payload, headers=headers)
    if response.status_code == 200:
        click.echo(response.json())
    elif response.status_code == 401:
        click.echo("Unauthorized: Invalid or missing authentication token.")
    else:
        click.echo(f"Error: {response.text}")

@click.command()
@click.option('--pattern', help='Basic pattern matching for scanning keys')
@click.option('--regex', help='Advanced regex scanning (writers only)')
@click.option('--limit', default=555, help='Limit for scanning results')
@click.option('--output-file', default=None, help='File to save the organized keys')
@click.option('-S', 'simple_output', is_flag=True, help='Output keys in plain format, one key per line')
def scan_keys(pattern, regex, limit, output_file, simple_output):
    """Scan keys based on a pattern or regex and group them by category."""
    params = {"limit": limit}
    if pattern:
        params["pattern"] = pattern
    if regex:
        params["regex"] = regex
    headers = {"Authorization": f"Bearer {EH_TOKEN}"}
    response = requests.get(f"{API_URL}/api/scan", params=params, headers=headers)
    if response.status_code == 200:
        keys = response.json().get('keys', [])
        if simple_output:
            for key in keys:
                click.echo(key)
        else:
            grouped_keys = group_keys_by_category(keys)
            display_grouped_keys(grouped_keys)
            if output_file:
                save_grouped_keys_to_file(grouped_keys, output_file)
                click.echo(f"Organized keys saved to {output_file}")
    elif response.status_code == 401:
        click.echo("Unauthorized: Invalid or missing authentication token.")
    else:
        click.echo(f"Error: {response.text}")

def group_keys_by_category(keys):
    grouped_keys = {}
    for key in keys:
        prefix = key.split(':')[0]
        if prefix not in grouped_keys:
            grouped_keys[prefix] = []
        grouped_keys[prefix].append(key)
    return grouped_keys

def display_grouped_keys(grouped_keys):
    for category, keys in grouped_keys.items():
        click.echo(f"{category}:")
        for key in keys:
            click.echo(f"  - {key}")

def save_grouped_keys_to_file(grouped_keys, output_file):
    with open(output_file, 'w') as f:
        for category, keys in grouped_keys.items():
            f.write(f"{category}:\n")
            for key in keys:
                f.write(f"  - {key}\n")

@click.command()
@click.option('--trace-id', required=True, help='Trace ID for visual replay')
def lesr_replay(trace_id):
    """Stream trace with echo session glyphs."""
    render_LESR_timeline(trace_id)
    click.echo(f"Trace {trace_id} replayed with echo session glyphs.")

def render_LESR_timeline(trace_id):
    """Render LESR timeline with glyphs, memory visuals, and tonal overlays."""
    click.echo(f"📡 Rendering LESR timeline for Trace ID: {trace_id}")

    # Simulate fetching session trace data
    session_data = {
        "glyph_stream": ["🔮", "✨", "🌌"],
        "memory_keys": ["key1", "key2", "key3"],
        "delta_map": {"modulation": "harmonic", "intensity": "medium"}
    }

    # Display glyph stream
    click.echo("\nGlyph Stream:")
    for glyph in session_data["glyph_stream"]:
        click.echo(f"  {glyph}")

    # Display memory key visuals
    click.echo("\nMemory Keys:")
    for key in session_data["memory_keys"]:
        click.echo(f"  - {key}")

    # Display tonal overlays
    click.echo("\nTonal Overlays:")
    for key, value in session_data["delta_map"].items():
        click.echo(f"  {key.capitalize()}: {value}")

    # Simulate animation feedback in terminal
    click.echo("\nAnimating feedback states...")
    for frame in ["|", "/", "-", "\\"]:
        click.echo(f"  {frame}", nl=False)
        click.pause(info="Simulating animation frame delay")

    click.echo("\nLESR timeline rendering complete.")

@click.command()
@click.option('--key', required=True, help='Memory key to retrieve and execute as a script')
@click.option('--verbose', is_flag=True, help='Enable verbose output for the executed script')
def run_memory_script(key, verbose):
    """Fetch a memory value by key and execute it as a Bash script."""
    try:
        # Use the internal get_memory function to fetch the memory content
        params = {"key": key}
        headers = {"Authorization": f"Bearer {EH_TOKEN}"}
        response = requests.get(f"{API_URL}/api/memory", params=params, headers=headers)

        if response.status_code != 200:
            click.echo(f"Error: Unable to fetch memory for key {key}. {response.text}")
            return

        # Extract the script from the JSON response
        script = response.json().get("value")
        if not script:
            click.echo(f"Error: No script found for key {key}")
            return

        # Execute the script with conditional verbosity
        if verbose:
            subprocess.run(script, shell=True)
        else:
            with open(os.devnull, 'w') as devnull:
                subprocess.run(script, shell=True, stdout=devnull, stderr=devnull)

        click.echo(f"Script executed successfully for key {key}.")
    except Exception as e:
        click.echo(f"An unexpected error occurred: {e}")

@click.command()
@click.option('--poll-interval', default=1.0, help='Polling interval in seconds')
@click.option('--ttl', default=300, help='Time to live in seconds')
@click.option('--verbose', is_flag=True, help='Enable verbose output')
def poll_clipboard_reflex(poll_interval, ttl, verbose):
    """Start the ReflexQL clipboard exchange polling loop."""
    start_time = time.time()
    
    def check_clipboard_flags():
        try:
            # Check if there's pending content
            params = {"key": "Reflex::Clipboard.Ready"}
            headers = {"Authorization": f"Bearer {EH_TOKEN}"}
            response = requests.get(f"{API_URL}/api/memory", params=params, headers=headers)
            
            if response.status_code == 200 and response.json().get("value") == "true":
                # Fetch pending content
                content_response = requests.get(
                    f"{API_URL}/api/memory",
                    params={"key": "Reflex::Clipboard.PendingContent"},
                    headers=headers
                )
                
                if content_response.status_code == 200:
                    content = content_response.json().get("value")
                    if content:
                        # Write to system clipboard using xclip
                        try:
                            process = subprocess.Popen(['xclip', '-selection', 'clipboard'], stdin=subprocess.PIPE)
                            process.communicate(input=content.encode())
                            
                            # Set delivered flag
                            requests.post(
                                f"{API_URL}/api/memory",
                                json={"key": "Reflex::Clipboard.Delivered", "value": "true"},
                                headers=headers
                            )
                            
                            if verbose:
                                click.echo("✨ Content delivered to clipboard")
                                
                        except FileNotFoundError:
                            click.echo("Error: xclip not found. Please install xclip package.")
                            return False
                        
        except Exception as e:
            if verbose:
                click.echo(f"Error in clipboard polling: {e}")
            return False
            
        return True

    click.echo("🌟 Starting ReflexQL clipboard polling loop...")
    
    while True:
        if not check_clipboard_flags():
            break
            
        # Check TTL
        if time.time() - start_time > ttl:
            click.echo("ReflexQL polling loop TTL expired")
            break
            
        time.sleep(poll_interval)

@click.group()
@click.option('--env', '-E', default=None, help='Path to an alternative environment file to load')
def cli(env):
    """Main CLI group with optional environment file loading."""
    if env:
        load_dotenv(env)
        click.echo(f"Loaded environment variables from {env}")

cli.add_command(scan_nodes)
cli.add_command(flex)
cli.add_command(trace_orbit)
cli.add_command(echo_sync)
cli.add_command(draw_memory_graph)
cli.add_command(curating_red_stones)
cli.add_command(activate_echonode_trace)
cli.add_command(enrich_fractale_version)
cli.add_command(graphbuilder_sync_command)
cli.add_command(redstone_encode_resonance)
cli.add_command(redstone_write_narrative_diffs)
cli.add_command(redstone_store_resonance_anchors)
cli.add_command(redstone_sync_echonode_metadata)
cli.add_command(redstone_post_commit_analysis)
cli.add_command(echo_live_reports)
cli.add_command(mia_status)
cli.add_command(tushell_echo)
cli.add_command(get_memory)
cli.add_command(post_memory)
cli.add_command(scan_keys)
cli.add_command(lesr_replay)
cli.add_command(run_memory_script)
cli.add_command(poll_clipboard_reflex)

if __name__ == '__main__':
    cli()
